#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-04-13
# @Filename: delegate.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Any, List

import numpy
from astropy.io import fits
from astropy.time import Time
from clu.legacy.types.pvt import PVT

from archon.actor.delegate import ExposureDelegate, FetchDataDict
from sdsstools.time import get_sjd

from .exceptions import SpecMechError


if TYPE_CHECKING:
    from clu.command import Command

    from yao.actor import YaoActor
    from yao.controller import YaoController


# TODO: LCOTCC cards are copied from flicamera. Should reunify code.


class YaoDelegate(ExposureDelegate["YaoActor"]):
    """Exposure delegate for BOSS."""

    def __init__(self, actor: YaoActor):
        super().__init__(actor)

        self.header_data = {}
        self.shutter_failed: bool = False

    def reset(self):
        """Reset the delegate."""

        self.shutter_failed = False

        return super().reset()

    async def pre_expose(self, controllers: List[YaoController]):
        """Runs the e-purge routine before the exposure."""

        self.command.info("E-purging and flushing.")

        await asyncio.gather(*[controller.purge() for controller in controllers])

        return True

    async def shutter(self, open: bool = False) -> bool:
        """Open/close the shutter."""

        expose_data = self.expose_data
        assert expose_data

        # TODO: this should be part of the ExposureDelegate.
        if expose_data.exposure_time == 0 or expose_data.flavour in ["bias", "dark"]:
            return True

        try:
            # Time-out in five seconds.
            await asyncio.wait_for(
                self.command.actor.spec_mech.pneumatic_move(
                    "shutter",
                    open=open,
                    command=self.command,
                ),
                5,
            )
        except asyncio.TimeoutError:
            self.command.error(
                "Timed out trying to move the shutter. The SpecMech may be dead."
            )
            self.shutter_failed = True
        except SpecMechError as err:
            self.command.error(f"Failed moving shutter: {err}")
            self.shutter_failed = True

        if self.shutter_failed:
            if open is True:
                # No point in exposing if we cannot open the shutter.
                return False

            # Try to read out with shutter open.
            self.command.warning("Reading out the exposure before failing.")

        return True

    async def readout(
        self,
        command: Command[YaoActor],
        extra_header: dict[str, Any] = {},
        delay_readout: int = 0,
        write: bool = True,
    ):
        """Reads detectors."""

        if self.shutter_failed:
            self.command.warning(
                "Frame was read out but shutter failed to close. "
                "There may be contamination in the image."
            )

        # Finish exposure tasks.
        try:
            if self._expose_cotasks:
                self.command.debug(text="Awaiting for exposure cotasks to finishing.")
                await asyncio.wait_for(self._expose_cotasks, 10)
        except asyncio.TimeoutError:
            self.command.warning("Timed out running exposure cotasks.")

        read_result = await super().readout(command, extra_header, delay_readout, write)

        return False if (self.shutter_failed or not read_result) else True

    async def expose_cotasks(self):
        """Grab header information during exposure."""

        self.command.debug(text="Starting exposure cotasks.")

        self.header_data = {}

        # FPS Cards
        self.header_data["fps_cards"] = [
            (
                "CONFID",
                get_keyword(
                    self.actor,
                    "jaeger",
                    "configuration_loaded",
                    idx=0,
                    cnv=int,
                ),
                "Configuration ID",
            ),
            (
                "DESIGNID",
                get_keyword(
                    self.actor,
                    "jaeger",
                    "configuration_loaded",
                    idx=1,
                    cnv=int,
                ),
                "Design ID associated with CONFIGID",
            ),
            (
                "FIELDID",
                get_keyword(
                    self.actor,
                    "jaeger",
                    "configuration_loaded",
                    idx=2,
                    cnv=int,
                ),
                "Field ID associated with CONFIGID",
            ),
        ]

        # TCC Cards
        self.header_data["lcotcc_cards"] = get_lcotcc_cards(self.actor)

        # Cherno/guider cards
        cherno_cards = []
        for idx, card in enumerate(["OFFRA", "OFFDEC", "OFFPA"]):
            default = get_keyword(
                self.actor,
                "cherno",
                "default_offset",
                idx=idx,
                default=0.0,
                cnv=float,
            )
            offset = get_keyword(
                self.actor,
                "cherno",
                "offset",
                idx=idx,
                default=0.0,
                cnv=float,
            )
            cherno_cards.append(
                (
                    card,
                    default + offset,
                    "Absolute guider offset in " + card.replace("OFF", ""),
                )
            )
        cherno_cards.append(
            (
                "SEEING",
                get_keyword(
                    self.actor,
                    "cherno",
                    "astrometry_fit",
                    idx=4,
                    cnv=float,
                ),
                "Seeing from the guider [arcsec]",
            )
        )
        self.header_data["cherno_cards"] = cherno_cards

        # Lamps/screen cards
        lamp_cards = []
        for lamp in ["TCS_FF", "Ne", "HeAr", "FF"]:
            value = get_keyword(self.actor, "lcolamps", lamp, idx=0, default="?")
            if value == "ON":
                card_value = "1 1 1 1"
            elif value == "OFF":
                card_value = "0 0 0 0"
            else:
                card_value = "? ? ? ?"

            if lamp == "TCS_FF":
                lamp_name = "FF"
            elif lamp == "FF":
                lamp_name = "M2FF"
            else:
                lamp_name = lamp.upper()

            lamp_cards.append((lamp_name, card_value, f"{lamp} lamps 1:On 0:Off"))

        screen_pos = float(get_keyword(self.actor, "lcotcc", "screenPos", 0, cnv=float))
        lamp_cards.append(("FFSPOS", screen_pos, "FF screen position [deg]"))

        # tel_alt = float(get_keyword(self.actor, "lcotcc", "axePos", 1, cnv=float))
        # if numpy.abs(tel_alt - screen_pos) < 40:  # HACK
        #     ffs_value = "1 1 1 1 1 1 1 1"
        # else:
        #     ffs_value = "0 0 0 0 0 0 0 0"
        # lamp_cards.append(("FFS", ffs_value, "Flatfield Screen 1:closed 0:open"))

        if self.expose_data and self.expose_data.flavour in ["flat", "arc"]:
            ffs_value = "1 1 1 1 1 1 1 1"
        else:
            ffs_value = "0 0 0 0 0 0 0 0"
        lamp_cards.append(("FFS", ffs_value, "Flatfield Screen 1:closed 0:open"))

        self.header_data["lamp_cards"] = lamp_cards

        # SpecMech cards
        specmech_cards = []

        status_left = await self.actor.spec_mech.pneumatic_status("left")
        status_right = await self.actor.spec_mech.pneumatic_status("right")
        if status_left == "closed" and status_right == "closed":
            hartmann = "Left,Right"
        elif status_left == "closed" and status_right == "open":
            hartmann = "Left"
        elif status_left == "open" and status_right == "closed":
            hartmann = "Right"
        elif status_left == "open" and status_right == "open":
            hartmann = "Out"
        else:
            hartmann = "?"
        specmech_cards.append(("HARTMANN", hartmann, "Hartmanns: Left,Right,Out"))

        for motor in ["a", "b", "c"]:
            _, pos, *_ = await self.actor.spec_mech.get_stat(f"motor-{motor}")
            specmech_cards.append(
                (
                    f"COLL{motor.upper()}",
                    int(pos),
                    f"The position of the {motor.upper()} collimator motor",
                )
            )

        # specMech data
        ori = [-999, -999, -999]
        try:
            ori = await self.actor.spec_mech.get_stat("orientation")
        except Exception as err:
            self.command.warning(f"Cannot get specMech orientation: {err}")
        finally:
            for ii, axis in enumerate(["X", "Y", "Z"]):
                specmech_cards.append(
                    (
                        f"MECHORI{axis}",
                        ori[ii],
                        f"Orientation in {axis} axis [cm/s2]",
                    )
                )

        mech_env = [-999] * 7
        try:
            mech_env = await self.actor.spec_mech.get_stat("environment")
        except Exception as err:
            self.command.warning(f"Cannot get specMech environment: {err}")
        finally:
            for ii, (label, comment) in enumerate(
                [
                    ("B2CAMT", "B2 camera temperature [degC]"),
                    ("B2CAMH", "B2 camera RH [%]"),
                    ("R2CAMT", "R2 camera temperature [degC]"),
                    ("R2CAMH", "R2 camera RH [%]"),
                    ("COLLT", "Collimator temperature [degC]"),
                    ("COLLH", "Collimator RH [%]"),
                    ("SPECMT", "specMech temperature [degC]"),
                ]
            ):
                specmech_cards.append((label, mech_env[ii], comment))

        self.header_data["specmech_cards"] = specmech_cards

        return await super().expose_cotasks()

    async def post_process(self, fdata: FetchDataDict):
        """Post-process images and rearrange overscan regions."""

        config = self.actor.config

        controller = self.actor.controllers[fdata["controller"]]

        LINES = controller.current_window["lines"]
        PIXELS = controller.current_window["pixels"]

        WIN_MODE = self.expose_data.window_mode if self.expose_data else None

        data = fdata["data"]
        header = fdata["header"]
        assert isinstance(data, numpy.ndarray)

        ccd = fdata["ccd"]
        if ccd not in ["b2", "r2"]:
            self.command.warning(text=f"Unknown CCD {ccd}.")
            return

        # Decide whether we want to do some callisthenics with the data to move
        # the overscan regions to the edges and to make things look exactly like
        # at APO.
        if config["controllers"]["sp2"].get("raw_mode", False) is False:
            if WIN_MODE != "hartmann":
                OL = config["controllers"]["sp2"]["overscan_regions"][ccd]["lines"]
                OL_END = -OL
            else:
                OL = 0
                OL_END = None

            OP = config["controllers"]["sp2"]["overscan_regions"][ccd]["pixels"]

            # Copy original data.
            rr = data.copy()

            # Rearrange pixel overscan.
            data[:, :OP] = rr[:, PIXELS - OP : PIXELS]
            data[:, -OP:] = rr[:, PIXELS : PIXELS + OP]

            # Rearrange line overscan.
            if OL > 0:
                data[:OL, OP:PIXELS] = rr[LINES - OL : LINES, : PIXELS - OP]  # BL
                data[-OL:, OP:PIXELS] = rr[LINES : LINES + OL, : PIXELS - OP]  # TL
                data[:OL, PIXELS:-OP] = rr[LINES - OL : LINES, PIXELS + OP :]  # BL
                data[-OL:, PIXELS:-OP] = rr[LINES : LINES + OL, PIXELS + OP :]  # BR

            # Rearrange data.
            data[OL:LINES, OP:PIXELS] = rr[: LINES - OL, : PIXELS - OP]  # BL
            data[LINES:OL_END, OP:PIXELS] = rr[LINES + OL :, : PIXELS - OP]  # TL
            data[OL:LINES, PIXELS:-OP] = rr[: LINES - OL, PIXELS + OP :]  # BR
            data[LINES:OL_END, PIXELS:-OP] = rr[LINES + OL :, PIXELS + OP :]  # TR

            if WIN_MODE == "hartmann":
                DEF_LINES = controller.default_window["lines"]
                DEF_PIXELS = controller.default_window["pixels"]
                new_data = numpy.zeros((DEF_LINES * 2, DEF_PIXELS * 2), dtype="u2")

                PRELINES = controller.current_window["preskiplines"]

                new_data[PRELINES : PRELINES + LINES, :] = data[:LINES, :]
                new_data[-PRELINES - LINES : -PRELINES, :] = data[LINES:, :]

                fdata["data"] = new_data
            else:
                fdata["data"] = data

        # Convert to astropy Header so it's easier to manipulate.
        header_ap = fits.Header()
        for key, value in header.items():
            if isinstance(value, (list, tuple)):
                header_ap[key] = tuple(value)
            else:
                header_ap[key] = value

        # Rename some keywords and add others to match APO BOSS datamodel.
        header_ap.insert("CCD", ("CAMERAS", header_ap["CCD"]))

        header_ap.rename_keyword("IMAGETYP", "FLAVOR")
        header_ap.rename_keyword("OBSTIME", "DATE-OBS")

        if self.expose_data and self.expose_data.flavour == "object":
            # SoS expects science.
            header_ap["FLAVOR"] = "science"

        header_ap["MJD"] = get_sjd()

        isot = Time(header_ap["DATE-OBS"], format="isot", scale="tai")
        tai_card = (
            "TAI-BEG",
            isot.mjd * 24 * 3600,
            "MJD(TAI) seconds at start of integration",
        )
        header_ap.insert("DATE-OBS", tai_card)

        reqtime_card = ("REQTIME", header_ap["EXPTIME"], "Requested exposure time")
        header_ap.insert("EXPTIME", reqtime_card)

        # Instrument cards
        header_ap.append(("CARTID", "FPS-S", "Instrument ID"))

        for card in self.header_data.get("fps_cards", []):
            header_ap.append(card)

        # TCC Cards
        for card in self.header_data.get("lcotcc_cards", []):
            header_ap.append(card)

        # Cherno offset cards
        for card in self.header_data.get("cherno_cards", []):
            header_ap.append(card)

        # Lamps
        for card in self.header_data.get("lamp_cards", []):
            header_ap.append(card)

        # specMech
        for card in self.header_data.get("specmech_cards", []):
            header_ap.append(card)

        # Convert back to dict of lists.
        update_header = {}
        for card in header_ap.cards:
            value = None if isinstance(card.value, fits.card.Undefined) else card.value
            update_header[card.keyword] = [value, card.comment]

        fdata["header"] = update_header


def pvt2pos(tup):
    pvt = PVT(*tup)
    return pvt.getPos()


def get_keyword(
    actor: YaoActor,
    model_name: str,
    key: str,
    idx: int | None = 0,
    default: Any = "NaN",
    cnv=None,
):
    """Returns the value of a keyword."""

    if not actor.tron or not actor.tron.models:
        actor.write("w", text=f"Cannot retrive keyword {key.upper()}.")
        return default

    model = actor.tron.models[model_name]

    try:
        value = model[key].value
        if idx is not None:
            value = value[idx]
        if cnv:
            value = cnv(value)

        # Headers cannot have NaN values as floats.
        try:
            if numpy.isnan(value):
                value = "NaN"
        except TypeError:
            pass

        return value
    except BaseException:
        return default


def get_lcotcc_cards(actor: YaoActor):
    """Return a list of cards describing the LCO TCC state."""

    model = "lcotcc"

    cards: list[tuple] = []

    objSysName = get_keyword(actor, "lcotcc", "objSys", 0, default="UNKNOWN")
    cards.append(("OBJSYS", objSysName, "The TCC objSys"))

    # ObjSys
    cards += [
        (
            "RA",
            get_keyword(actor, model, "objNetPos", 0, cnv=pvt2pos),
            "RA of telescope boresight (deg)",
        ),
        (
            "DEC",
            get_keyword(actor, model, "objNetPos", 1, cnv=pvt2pos),
            "Dec of telescope boresight (deg)",
        ),
        (
            "RADEG",
            get_keyword(actor, model, "objNetPos", 0, cnv=pvt2pos),
            "RA of telescope pointing (deg)",
        ),
        (
            "DECDEG",
            get_keyword(actor, model, "objNetPos", 1, cnv=pvt2pos),
            "Dec of telescope pointing (deg)",
        ),
        (
            "Equinox",
            2000.0,
            "Equinox of celestial coordinate system",
        ),
        (
            "AZ",
            get_keyword(actor, model, "axePos", 0, cnv=float),
            "Azimuth axis pos. (approx, deg)",
        ),
        (
            "ALT",
            get_keyword(actor, model, "axePos", 1, cnv=float),
            "Altitude axis pos. (approx, deg)",
        ),
        (
            "AIRMASS",
            get_keyword(actor, model, "airmass", 0, cnv=float),
            "Airmass",
        ),
    ]

    # Rotator
    cards.append(
        (
            "HA",
            get_keyword(actor, model, "tccHA", 0, cnv=float),
            "HA axis pos. (approx, deg)",
        )
    )

    cards.append(
        (
            "ROTPOS",
            get_keyword(actor, model, "axePos", 2, cnv=float),
            "Rotator requested pos. (approx, deg)",
        )
    )

    cards.append(
        (
            "IPA",
            get_keyword(actor, model, "axePos", 2, cnv=float),
            "Rotator axis pos. (approx, deg)",
        )
    )

    # Focus / M2
    cards.append(
        (
            "FOCUS",
            get_keyword(actor, model, "secFocus", 0, cnv=float),
            "User-specified focus offset (um)",
        )
    )

    orient_names = ("piston", "xtilt", "ytilt", "xtran", "ytran", "zrot")
    for ii in range(len(orient_names)):
        cards.append(
            (
                "M2" + orient_names[ii].upper(),
                get_keyword(actor, model, "secOrient", ii, cnv=float),
                "TCC SecOrient",
            )
        )

    # Temperatures
    cards.append(
        (
            "T_OUT",
            get_keyword(actor, model, "tccTemps", 0, cnv=float),
            "Outside temperature deg C.",
        )
    )

    cards.append(
        (
            "T_IN",
            get_keyword(actor, model, "tccTemps", 1, cnv=float),
            "Inside temperature deg C.",
        )
    )

    cards.append(
        (
            "T_PRIM",
            get_keyword(actor, model, "tccTemps", 2, cnv=float),
            "Primary mirror temperature deg C.",
        )
    )

    cards.append(
        (
            "T_CELL",
            get_keyword(actor, model, "tccTemps", 3, cnv=float),
            "Cell temperature deg C.",
        )
    )

    cards.append(
        (
            "T_FLOOR",
            get_keyword(actor, model, "tccTemps", 4, cnv=float),
            "Floor temperature deg C.",
        )
    )

    cards.append(
        (
            "T_TRUSS",
            get_keyword(actor, "lcotcc", "secTrussTemp", 0, cnv=float),
            "Truss temperature deg C.",
        )
    )

    return cards



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-10-22
# @Filename: alerts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio
import time
from collections import deque

from typing import TYPE_CHECKING, Type

from yao import config


if TYPE_CHECKING:
    from .actor import YaoActor

__all__ = ["AlertsBot"]


class AlertsBot:
    """Handles alerts for BOSS."""

    def __init__(self, actor: YaoActor):
        __ALERTS__: list[Type[BaseAlert]] = [
            R2CCDTemperatureAlert,
            R2LN2TemperatureAlert,
            B2CCDTemperatureAlert,
            B2LN2TemperatureAlert,
            YaoHeartbeat,
        ]

        self.actor = actor

        self.alerts = [AlertClass(self.actor).start() for AlertClass in __ALERTS__]

    def stop(self):
        """Stops all the alert handlers."""

        for alert in self.alerts:
            alert.stop()


class BaseAlert(metaclass=abc.ABCMeta):
    """An alert handler."""

    name: str
    keyword: str

    def __init__(self, actor: YaoActor, interval: float = 60.0):
        self.actor = actor
        self.interval = interval

        self._value: bool = False

        if not hasattr(self, "name"):
            self.name = self.__class__.__name__

        if not hasattr(self, "keyword"):
            raise NameError(f"Alert {self.name} has no defined keyword.")

        self._task: asyncio.Task | None = None

        self.reset()

    def reset(self):
        """Resets the alarm."""

        self._value = False
        self._notify(False, force=True)

    @property
    def value(self):
        """The value of the alert."""

        return self._value

    @value.setter
    def value(self, new_value: bool):
        """Sets the new value of the alert and notifies."""

        self._notify(new_value, force=False)
        self._value = new_value

    def _notify(self, value: bool, force: bool = False):
        """Notifies if the alert is on or off."""

        # Output positive alerts every time, but false ones only if they
        # change the current value of the alert.
        if value == self.value and value is False and not force:
            return

        code = "w" if value is True else "i"
        self.actor.write(code, {self.keyword: int(value)})

    def start(self):
        """Starts the alert handler."""

        if self._task is not None:
            self.stop()

        self._task = asyncio.create_task(self._alert_handler())

        return self

    def stop(self):
        """Stops the alert handler."""

        if self._task:
            self._task.cancel()

        self._task = None

    async def _alert_handler(self):
        """Execute the alert code in a loop."""

        await asyncio.sleep(1)

        while True:
            try:
                await self.check_alert()
            except Exception as err:
                self.on_error(err)

            await asyncio.sleep(self.interval)

    @abc.abstractmethod
    async def check_alert(self):
        """Code that actually checks and raises an alert."""

        pass

    def on_error(self, error: Exception):
        """Executed when an error is raised while checking the alert."""

        msg = f"Failed checking alert {self.name!r}"
        if str(error) != "":
            msg += f": {error}"
        else:
            msg += "."

        self.actor.write("w", {"error": msg})


class TemperatureAlert(BaseAlert):
    controller: str
    ccd: str
    status_param: str
    setpoint_param: str | float
    max_increase: float

    def __init__(self, actor: YaoActor, interval: float = 60):
        # Rolling log of alert values; keeps only the last two measurements.
        self._values_log = deque(maxlen=2)

        super().__init__(actor, interval)

    async def check_alert(self):
        """Checks the CCD temperature."""

        controller = self.actor.controllers[self.controller]

        status = await controller.get_device_status()

        temperature = status[self.status_param]

        if isinstance(self.setpoint_param, str):
            if not controller.acf_config:
                raise ValueError("Controller has no configuration loaded.")
            setpoint = controller.acf_config["CONFIG"].getfloat(self.setpoint_param)
        else:
            setpoint = self.setpoint_param

        if temperature > setpoint + self.max_increase:
            self._values_log.append(True)
        else:
            self._values_log.append(False)

        # We require two consecutive measurements to change the alarm.
        if len(self._values_log) < 2:
            pass
        elif self._values_log[0] is True and self._values_log[1] is True:
            self.value = True
        elif self._values_log[0] is False and self._values_log[1] is False:
            self.value = False

    def on_error(self, error: Exception):
        """Executed when an error is raised while checking the alert."""

        super().on_error(error)
        self._values_log.append(True)
        self.value = True


class R2LN2TemperatureAlert(TemperatureAlert):
    """r2 LN2 alert."""

    name = "r2_ln2_temp_alert"
    keyword = "r2_ln2_temp_alert"

    controller = "sp2"
    ccd = "r2"
    status_param = config["alerts"]["sp2"]["r2_ln2_status_param"]
    setpoint_param = config["alerts"]["sp2"]["ln2_base_temperature"]
    max_increase = config["alerts"]["sp2"]["ln2_temperature_max_increase"]


class R2CCDTemperatureAlert(TemperatureAlert):
    """r2 CCD alert."""

    name = "r2_ccd_temp_alert"
    keyword = "r2_ccd_temp_alert"

    controller = "sp2"
    ccd = "r2"
    status_param = config["alerts"]["sp2"]["r2_ccd_status_param"]
    setpoint_param = config["alerts"]["sp2"]["r2_setpoint_param"]
    max_increase = config["alerts"]["sp2"]["ccd_temperature_max_increase"]


class B2LN2TemperatureAlert(TemperatureAlert):
    """b2 LN2 alert."""

    name = "b2_ln2_temp_alert"
    keyword = "b2_ln2_temp_alert"

    controller = "sp2"
    ccd = "b2"
    status_param = config["alerts"]["sp2"]["b2_ln2_status_param"]
    setpoint_param = config["alerts"]["sp2"]["ln2_base_temperature"]
    max_increase = config["alerts"]["sp2"]["ln2_temperature_max_increase"]


class B2CCDTemperatureAlert(TemperatureAlert):
    """b2 CCD alert."""

    name = "b2_ccd_temp_alert"
    keyword = "b2_ccd_temp_alert"

    controller = "sp2"
    ccd = "b2"
    status_param = config["alerts"]["sp2"]["b2_ccd_status_param"]
    setpoint_param = config["alerts"]["sp2"]["b2_setpoint_param"]
    max_increase = config["alerts"]["sp2"]["ccd_temperature_max_increase"]


class YaoHeartbeat(BaseAlert):
    """Heartbeat alert."""

    name = "yao_heartbeat"
    keyword = "alive_at"

    reset_on_start = False

    def _notify(self, *args, **kwargs):
        """Disable normal alert notifications for this alert."""

        pass

    async def check_alert(self):
        """Reports if the actor is alive."""

        self.actor.write("d", {self.keyword: time.time()}, broadcast=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Aidan Gray (Aidan.Gray@idg.jhu.edu), José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-05-26
# @Filename: mech_controller.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
import re
import warnings

from typing import TYPE_CHECKING, Any

from sdsstools.logger import SDSSLogger

from yao import config

from .exceptions import SpecMechError, YaoUserWarning


if TYPE_CHECKING:
    from .actor import YaoCommand


__all__ = ["MechController", "ReplyCode", "SpecMechReply", "STATS", "check_reply"]


#: Stat names accepted by the report command.
STATS: dict[str, str] = {
    "time": "rt",
    "version": "rV",
    "environment": "re",
    "vacuum": "rv",
    "motors": "rd",
    "motor-a": "ra",
    "motor-b": "rb",
    "motor-c": "rc",
    "orientation": "ro",
    "pneumatics": "rp",
    # "nitrogen": "rn",
    "specmech": "rs",
}


def check_reply(reply: SpecMechReply):
    """Checks a specMech reply."""

    if reply.code == ReplyCode.ERR_IN_REPLY:
        for reply_data in reply.data:
            if "ERR" in reply_data[0]:
                code, msg = reply_data[1:]
                raise SpecMechError(f"Error {code} found in specMech reply: {msg!r}.")

    if reply.code == ReplyCode.CONTROLLER_REBOOTED:
        raise SpecMechError(
            "The specMech controller has rebooted. "
            "Acknowledge the reboot before continuing."
        )

    if reply.code == ReplyCode.CONNECTION_FAILED:
        raise SpecMechError("The connection to the specMech failed. Try reconnecting.")

    if reply.code != ReplyCode.VALID and reply.code != ReplyCode.REBOOT_ACKNOWLEDGED:
        raise SpecMechError(f"Failed parsing specMech reply: {reply.code.name!r}.")


class ReplyCode(enum.Enum):
    """Reply error codes."""

    VALID = enum.auto()
    UNPARSABLE_RESPONSE = enum.auto()
    MISMATCHED_COMMAND_ID = enum.auto()
    INVALID_COMMAND_CHECKSUM = enum.auto()
    INVALID_REPLY_CHECKSUM = enum.auto()
    ERR_IN_REPLY = enum.auto()
    CONTROLLER_REBOOTED = enum.auto()
    REBOOT_ACKNOWLEDGED = enum.auto()
    CONNECTION_FAILED = enum.auto()


class SpecMechReply:
    """A valid response to a command to the specMech."""

    def __init__(self, raw: bytes):
        self.command_id = 0
        self.raw = raw

        self.code = ReplyCode.VALID

        self.reply_sentence: str = ""
        self.data: list[list[str]] = []

        self.parse()

    def __str__(self):
        return self.raw

    def __repr__(self):
        return f"<MechControlReply (raw={self.raw!r})>"

    @staticmethod
    def calculate_checksum(message: bytes):
        """Computes the checksum field for the NMEA protocol.

        The checksum is simple, just an XOR of all the bytes between the
        ``$`` and the ``*`` (not including the delimiters themselves),
        and written in hexadecimal.

        """

        match = re.match(rb"^\$?(?:(.*?)(?:\*(?:[0-9A-Za-z]+)?)?)$", message)
        if match is None:
            raise ValueError(f"Unparsable message {message!r}")

        message = match.group(1)

        checksum = 0
        for b in message:
            checksum ^= b

        return f"{checksum:02X}"

    @staticmethod
    def check_checksum(message: bytes):
        """Checks the checksum from a reply.

        Parameters
        ----------
        message
            The message to check. Must include the checksum ``*XX``.

        """

        match = re.match(rb"^(.+?)\*([0-9A-Fa-f]+)$", message, re.DOTALL)

        if match is None:
            raise ValueError(f"Cannot parse checksum in message {message!r}")

        data, checksum = match.groups()

        return checksum.decode() == SpecMechReply.calculate_checksum(data)

    def parse(self):
        """Parses the raw reply."""

        # Check if the the response indicates a reboot. ! only ever happens if
        # a reboot has occurred.
        if b"!" in self.raw:
            self.code = ReplyCode.CONTROLLER_REBOOTED
            return

        # Get only the part without telnet subnegotiations and without the
        # terminator.
        match1 = re.match(b"(?:\xff.+\xf0)?(.*?)\x00?\n?>", self.raw, re.DOTALL)

        if match1 is None or len(match1.groups()) != 1:
            self.code = ReplyCode.UNPARSABLE_RESPONSE
            return

        data = match1.group(1)

        # Check if the the response indicates a reboot.
        if data == b"":
            self.code = ReplyCode.REBOOT_ACKNOWLEDGED
            return

        match2 = re.match(rb"(\$S2CMD.+?)\r(?:\x00?\n(.+)\r)?", data, re.DOTALL)

        if match2 is None or len(match2.groups()) != 2:
            self.code = ReplyCode.UNPARSABLE_RESPONSE
            return

        cmd, replies = match2.groups()

        if not self.check_checksum(cmd):
            self.code = ReplyCode.INVALID_COMMAND_CHECKSUM
            return

        match_command_id = re.match(rb".+?;([0-9])+\*", cmd)
        if match_command_id is None:
            warnings.warn(
                f"Failed matching command ID in command echo {cmd!r}",
                YaoUserWarning,
            )
        else:
            self.command_id = int(match_command_id.group(1))

        # The reply is only an echo of the command, without additional data.
        if replies is None:
            return

        reply_list = re.split(b"\r\x00\n", replies)

        for reply in reply_list:
            if not self.check_checksum(reply):
                self.code = ReplyCode.INVALID_REPLY_CHECKSUM
                return

            match3 = re.match(rb"^\$S2([A-Za-z0-9]+),(.+?)\*[0-9A-Fa-f]+$", reply)

            if match3 is None or len(match2.groups()) != 2:
                self.code = ReplyCode.UNPARSABLE_RESPONSE
                return

            sentence, data_pack = match3.groups()
            if sentence == b"ERR":
                self.code = ReplyCode.ERR_IN_REPLY

            # The data includes the sentence and the timestamp.
            self.data.append([sentence.decode()] + data_pack.decode().split(","))


class MechController:
    """Controller for the spectrograph mechanics.

    The `.MechController` handles the connection to the spectrograph
    mechanics microcontroller. A description of the communication
    protocol is available `here <https://bit.ly/38Xn2VE>`__.

    Parameters
    ----------
    address
        The IP address or host of the mech server.
    port
        The port of the host of the mech server.
    log
        A logger to which to write.

    """

    def __init__(self, address: str, port: int = 23, log: SDSSLogger | None = None):
        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None

        self.reboot: bool = False
        self.command_number: int = 0

        self.log = log or SDSSLogger("boss-spech-mech-client")

        self.spechMechAddress = address
        self.specMechPort = port

        self.lock = asyncio.Lock()

    async def start(self):
        """Opens a connection with the given IP and port."""

        self.log.info(
            f"Opening connection with {self.spechMechAddress} "
            f"on port {self.specMechPort}"
        )

        loop = asyncio.get_running_loop()
        loop.set_exception_handler(self.log.asyncio_exception_handler)

        connect = asyncio.open_connection(self.spechMechAddress, self.specMechPort)
        self.reader, self.writer = await asyncio.wait_for(connect, timeout=3)

    def is_connected(self):
        """Checks if we are connected to the specMech."""

        if not self.writer:
            return False

        return not self.writer.is_closing()

    async def send_data(self, command: str, timeout: float | None = None):
        """Sends the given string to the specMech and then awaits a response.

        Currently when a command is sent the controller is locked and any
        new command is blocked until a reply for the currenly running command
        arrives or a timeout happens.

        Parameters
        ----------
        command
            A string that is sent to specMech.
        timeout
            How long to wait for replies.

        Returns
        -------
        replies
            A tuple in which the first element is bytes array with the raw reply,
            and successive items are tuples with the data associated with each
            reply. If a timeout occurred before

        """

        # Only send '!\r' to acknowledge a reboot
        if command == "!":
            commandFinal = command + "\r"
        else:
            # Increment command number
            self.command_number += 1

            # Add command identifier to the command
            commandFinal = command + ";" + str(self.command_number) + "\r"

        # Send the command
        self.log.debug(f"Sent to specMech: {commandFinal!r}")

        try:
            if self.writer is None:
                raise ConnectionResetError("SpecMech client not connected.")

            async with self.lock:
                self.writer.write(commandFinal.encode())
                reply = await asyncio.wait_for(self.read_data(), timeout)

        except ConnectionResetError:
            reply = SpecMechReply(b"")
            reply.code = ReplyCode.CONNECTION_FAILED

            self.writer = None
            self.reader = None

        return reply

    async def read_data(self):
        """Awaits responses from specMech until the EOM character '>' is seen.

        Returns
        -------
        mech_reply
            A `.SpecMechReply` object with the reply received.

        """

        if self.reader is None:
            raise ConnectionResetError("SpecMech client not connected.")

        dataRaw = await self.reader.read(1024)

        # Continue accepting responses until '>' is received
        while b">" not in dataRaw and b"!" not in dataRaw:
            dataRawTmp = await self.reader.read(1024)
            dataRaw = dataRaw + dataRawTmp

        self.log.debug(f"Received from specMech: {dataRaw!r}")

        reply = SpecMechReply(dataRaw)

        if reply.code == ReplyCode.CONTROLLER_REBOOTED:
            self.reboot = True
        else:
            self.reboot = False

        return reply

    async def close(self):
        """Closes the connection with specMech."""

        self.log.info("Closing the connection to the specMech.")

        if self.writer is not None:
            self.writer.close()
            await self.writer.wait_closed()

        self.writer = None
        self.reader = None

    async def get_stat(self, stat: str) -> tuple[Any, ...]:
        """Returns the output of the report commands.

        Parameters
        ----------
        stat
            The report stat to recover. One of `.STATS`.

        Returns
        -------
        reply
            The parsed command reply values.

        """

        stat = stat.lower()

        if stat in STATS:
            mech_command = STATS[stat]
        else:
            raise SpecMechError(f"Invalid specMech stat {stat!r}.")

        reply = await self.send_data(mech_command)
        check_reply(reply)

        values = reply.data[0]

        if values[0] == "MTR":
            if stat.startswith("motor-"):
                # Only one reply.
                mtr = values[2]
                mtrPosition = int(values[3])
                mtrSpeed = int(values[5])
                mtrCurrent = int(values[7])
                mtrDirection = values[9]
                mtrLimit = True if values[11] == "Y" else False
                return (mtr, mtrPosition, mtrSpeed, mtrCurrent, mtrDirection, mtrLimit)
            elif stat == "motors":
                # This is the only case in which we have multiple replies.
                # Return only the positions as floats.
                positions = []
                for d in reply.data:
                    positions.append(int(d[3]))
                return tuple(positions)

            else:
                raise SpecMechError("Invalid stat for MTR reply.")

        elif values[0] == "ENV":
            env0T = float(values[2])
            env0H = float(values[4])
            env1T = float(values[6])
            env1H = float(values[8])
            env2T = float(values[10])
            env2H = float(values[12])
            specMechT = float(values[14])
            return (env0T, env0H, env1T, env1H, env2T, env2H, specMechT)

        elif values[0] == "ORI":
            accx = float(values[2])
            accy = float(values[3])
            accz = float(values[4])
            return (accx, accy, accz)

        elif values[0] == "PNU":
            # change the c/o/t and 0/1 responses of specMech
            # to something more readable
            if values[2] == "c":
                pnus = "closed"
            elif values[2] == "o":
                pnus = "open"
            else:
                pnus = "transiting"

            if values[4] == "c":
                pnul = "closed"
            elif values[4] == "o":
                pnul = "open"
            else:
                pnul = "transiting"

            if values[6] == "c":
                pnur = "closed"
            elif values[6] == "o":
                pnur = "open"
            else:
                pnur = "transiting"

            if values[8] == "0":
                pnup = "off"
            else:
                pnup = "on"

            return (pnus, pnul, pnur, pnup)

        elif values[0] == "TIM":
            tim = values[1]
            stim = values[2]
            btm = values[4]
            return (btm, tim, stim)

        elif values[0] == "VER":
            ver = values[2]
            return (ver,)

        elif values[0] == "VAC":
            red = float(values[2])
            blue = float(values[4])
            return (red, blue)

        elif values[0] == "LN2":
            valves = []
            for valve_status in values[2]:
                if valve_status.upper() == "C":
                    valves.append("closed")
                elif valve_status.upper() == "O":
                    valves.append("open")
                elif valve_status.upper() == "T":
                    valves.append("timeout")
                elif valve_status.upper() == "X":
                    valves.append("disabled")
                else:
                    valves.append("?")

            (
                buffer_dewar_supply_status,
                buffer_dewar_vent_status,
                red_dewar_vent_status,
                blue_dewar_vent_status,
            ) = valves

            time_next_fill = int(values[3])
            max_valve_open_time = int(values[5])
            fill_interval = int(values[7])
            ln2_pressure = int(values[9])

            if values[11].upper() == "C":
                buffer_dewar_thermistor_status = "cold"
            elif values[11].upper() == "H":
                buffer_dewar_thermistor_status = "warm"
            else:
                buffer_dewar_thermistor_status = "?"

            if values[13].upper() == "C":
                red_dewar_thermistor_status = "cold"
            elif values[13].upper() == "H":
                red_dewar_thermistor_status = "warm"
            else:
                red_dewar_thermistor_status = "?"

            if values[15].upper() == "C":
                blue_dewar_thermistor_status = "cold"
            elif values[15].upper() == "H":
                blue_dewar_thermistor_status = "warm"
            else:
                blue_dewar_thermistor_status = "?"

            return (
                buffer_dewar_supply_status,
                buffer_dewar_vent_status,
                red_dewar_vent_status,
                blue_dewar_vent_status,
                time_next_fill,
                max_valve_open_time,
                fill_interval,
                ln2_pressure,
                buffer_dewar_thermistor_status,
                red_dewar_thermistor_status,
                blue_dewar_thermistor_status,
            )

        elif "specmech":
            fan = "on" if int(values[2]) else "off"
            volts = float(values[4])
            return (fan, volts)

        else:
            raise SpecMechError(f"Invalid reply sentence {values[0]}.")

    async def pneumatic_move(
        self,
        mechanism: str,
        open: bool = True,
        command: YaoCommand | None = None,
    ):
        """Opens/closes a pneumatic mechanism.

        Parameters
        ----------
        mechanism
            Either ``shutter``, ``left``, or ``right``.
        open
            If `True`, opens the mechanism, otherwise closes it.
        command
            An actor command for outputs.

        """

        if mechanism == "left":
            spec_command = "ol" if open else "cl"
        elif mechanism == "right":
            spec_command = "or" if open else "cr"
        elif mechanism == "shutter":
            spec_command = "os" if open else "cs"
        else:
            raise SpecMechError(f"Invalid mechanism {mechanism!r}.")

        reply = await self.send_data(spec_command)
        check_reply(reply)

        # Check that all the mechanisms have arrived to their desired position.
        # Try twice, then fail.

        for ii in [1, 2]:
            timeout = config["timeouts"]["pneumatics"]
            if ii == 2:
                # Longer timeout if the first one failed.
                timeout *= 3

            await asyncio.sleep(timeout)

            reached = True

            status = await self.send_data("rp")
            try:
                check_reply(status)
            except SpecMechError as err:
                raise SpecMechError(
                    "Failed checking the status of the "
                    f"pneumatics after a move: {err}"
                )

            if mechanism == "shutter":
                mech_position = status.data[0][2]
            elif mechanism == "left":
                mech_position = status.data[0][4]
            elif mechanism == "right":
                mech_position = status.data[0][6]
            else:
                continue

            destination = "open" if open else "closed"
            if mech_position != destination[0]:
                reached = False

            if reached is True:
                if command:
                    mech_key = mechanism
                    if mechanism in ["left", "right"]:
                        mech_key = "hartmann_" + mech_key
                    command.info(message={mech_key: destination})
                return True

            if ii == 1:
                if command:
                    command.warning(
                        "Pneumatics did not reach the desired position. "
                        "Waiting a bit longer ..."
                    )
            else:
                if command:
                    await command.send_command("yao", "mech status pneumatics")
                raise SpecMechError("Pneumatics did not reach the desired position.")

        # We should never get here.
        raise SpecMechError

    async def pneumatic_status(self, mechanism: str) -> str:
        """Returns the open/closed status of a mechanism."""

        mechanism = mechanism.lower()
        if mechanism not in ["shutter", "left", "right"]:
            raise SpecMechError(f"Invalid mechanism {mechanism!r}.")

        status = await self.get_stat("pneumatics")

        if mechanism == "shutter":
            return status[0]
        elif mechanism == "left":
            return status[1]
        else:
            return status[2]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-04-01
# @Filename: exceptions.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


class YaoError(Exception):
    """A custom core Yao exception"""

    def __init__(self, message=None):
        message = "There has been an error" if not message else message
        super(YaoError, self).__init__(message)


class YaoNotImplemented(YaoError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):
        message = "This feature is not implemented yet." if not message else message
        super(YaoNotImplemented, self).__init__(message)


class SpecMechError(YaoError):
    """A specMech error."""


class YaoMissingDependency(YaoError):
    """A custom exception for missing dependencies."""

    pass


class YaoWarning(Warning):
    """Base warning for Yao."""

    pass


class YaoUserWarning(UserWarning, YaoWarning):
    """The primary warning class."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-04-01
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import os
import pathlib
import warnings

from clu import Command
from clu.legacy import LegacyActor

from archon import log as archon_log
from archon.actor.actor import ArchonBaseActor
from archon.actor.tools import get_schema

from yao import __version__, config
from yao.alerts import AlertsBot
from yao.commands import parser
from yao.controller import YaoController
from yao.delegate import YaoDelegate
from yao.exceptions import YaoUserWarning
from yao.mech_controller import MechController


class YaoActor(ArchonBaseActor, LegacyActor):
    """Yao actor."""

    is_legacy = True

    DELEGATE_CLASS = YaoDelegate
    CONTROLLER_CLASS = YaoController

    parser = parser

    def __init__(self, *args, **kwargs):
        self._message_processor = self._process_message

        schema = self.merge_schemas(kwargs.pop("schema", None))

        # We are going to append "spX_" to every keyword that contains the
        # controller sub-keyword, so for now let's accept additionalProperties.
        schema["additionalProperties"] = True

        super().__init__(*args, schema=schema, **kwargs)

        self.version = __version__

        self.alerts_bot: AlertsBot | None = None

        # TODO: this assumes one single mech controller, not one per spectrograph,
        # but in practice for now that's fine.
        self.spec_mech = MechController(
            self.config["specMech"]["address"],
            self.config["specMech"]["port"],
        )

        # Add actor log handlers to the archon library to also get that logging.
        archon_log.addHandler(self.log.sh)
        if self.log.fh:
            archon_log.addHandler(self.log.fh)
            self.log.fh.setLevel(10)

    async def start(self):
        """Starts the actor and connects the specMech client."""

        try:
            await self.spec_mech.start()
        except asyncio.TimeoutError:
            warnings.warn(
                "Failed connecting to mech controller: timed out.",
                YaoUserWarning,
            )
        except Exception as err:
            warnings.warn(
                f"Failed connecting to mech controller: {err}",
                YaoUserWarning,
            )

        new_self = await super().start()

        if self.alerts_bot is None:
            self.alerts_bot = AlertsBot(self)

        return new_self

    def merge_schemas(self, yao_schema_path: str | None = None):
        """Merge default schema with the one from yao."""

        schema = get_schema()  # Default archon schema.

        if yao_schema_path:
            root_path = pathlib.Path(__file__).absolute().parent
            if not os.path.isabs(yao_schema_path):
                yao_schema_path = os.path.join(str(root_path), yao_schema_path)

            yao_schema = json.loads(open(yao_schema_path, "r").read())

            schema["definitions"].update(yao_schema.get("definitions", {}))
            schema["properties"].update(yao_schema.get("properties", {}))
            schema["patternProperties"].update(yao_schema.get("patternProperties", {}))

            if "additionalProperties" in yao_schema:
                schema["additionalProperties"] = yao_schema["additionalProperties"]

        return schema

    @staticmethod
    def _process_message(message: dict):
        """Processes the messages to output to the users.

        If the message is a dictionary with a ``controller`` keyword,
        reformats it into multiple keywords that start with the controller
        name.

        """

        original_message = message.copy()
        new_message = {}

        for key, value in original_message.items():
            if not isinstance(value, dict):
                new_message[key] = value
                continue

            if "controller" in value:
                controller_prefix = value["controller"] + "_"
            else:
                controller_prefix = ""

            for subkey, subvalue in value.items():
                if subkey == "controller":
                    continue
                if isinstance(subvalue, dict):
                    subvalue = list(subvalue.values())
                subkey = subkey.replace("/", "_")
                new_message[controller_prefix + subkey] = subvalue

        return new_message

    @classmethod
    def from_config(cls, file: str | None = None):
        """Creates an actor from the internal package configuration."""

        return super().from_config(config if file is None else file)


YaoCommand = Command[YaoActor]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Aidan Gray (Aidan.Gray@idg.jhu.edu), José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-05-26
# @Filename: commands.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import re

from typing import TYPE_CHECKING

import click
import numpy
import numpy.typing

from archon.actor import parser

from yao.exceptions import SpecMechError
from yao.mech_controller import STATS, ReplyCode, check_reply


if TYPE_CHECKING:
    from yao.mech_controller import SpecMechReply

    from .actor import YaoCommand


__all__ = ["mech"]


async def check_controller(command: YaoCommand) -> bool:
    """Performs sanity check in the controller.

    Outputs error messages if a problem is found. Return `False` if the controller
    is not in a valid state.

    """

    mech = command.actor.spec_mech

    if not mech.is_connected():
        command.fail(error="specMech controller not connected.")
        return False

    try:
        await mech.send_data("rt", timeout=3)
    except asyncio.TimeoutError:
        command.fail(error="specMech timed out replying.")
        return False

    return True


def parse_reply(command: YaoCommand, reply: SpecMechReply, fail: bool = True):
    """Parses a reply from the mech controller.

    If the reply indicates an error or that the controller has been rebooted,
    fails the command with an appropriate message and returns `False`.

    """

    if fail:
        write_func = command.fail
    else:
        write_func = command.error

    try:
        check_reply(reply)
    except SpecMechError as err:
        write_func(str(err))
        return False

    return True


@parser.group()
def mech(*args):
    """Interface to the specMech controller."""

    pass


# Values to output if a stat fails.
FAILED_STAT_REPLIES = {
    "nitrogen": ["?", "?", "?", "?", -999, -999, -999, -999, "?", "?", "?", "?"],
    "motor-a": ["a", -999, -999, -999, -999, False],
    "motor-b": ["b", -999, -999, -999, -999, False],
    "motor-c": ["c", -999, -999, -999, -999, False],
    "motors": [-999, -999, -999],
    "environment": [-999.0] * 7,
    "orientation": [-999.0] * 3,
    "pneumatics": ["?"] * 4,
    "time": ["?"] * 3,
    "version": ["?"],
    "vacuum": [-999.0, -999.0],
    "specmech": ["?", -999.0],
}


@mech.command()
@click.argument("STAT", type=click.Choice(list(STATS)), required=False)
@click.option("-d", "--debug", is_flag=True, help="Uses debug status in outputs..")
async def status(
    command: YaoCommand,
    controllers,
    stat: str | None = None,
    debug: bool = False,
):
    """Queries specMech for all status responses."""

    if not (await check_controller(command)):
        return

    if stat is None:
        process_stats = list(STATS)
    else:
        process_stats = [stat.lower()]

    if debug:
        write_func = command.debug
    else:
        write_func = command.info

    for this_stat in process_stats:
        if this_stat not in STATS:
            return command.fail(f"Invalid specMech status command {this_stat!r}.")

        mech = command.actor.spec_mech

        try:
            values = await mech.get_stat(this_stat)
        except SpecMechError as err:
            command.error(str(err))
            if this_stat in FAILED_STAT_REPLIES:
                values = FAILED_STAT_REPLIES[this_stat]
            else:
                continue

        if this_stat.startswith("motor-"):
            pos, vel, current, _, limit = values[1:]
            write_func(
                message={this_stat.replace("-", "_"): [pos, vel, current, limit]}
            )

            if limit is True:
                motor = this_stat[-1].upper()
                command.warning(f"Motor {motor}: limit switch has been triggered!")

        elif this_stat == "motors":
            write_func(motor_positions=values)

        elif this_stat == "environment":
            write_func(
                temperature_blue=values[0],
                humidity_blue=values[1],
                temperature_red=values[2],
                humidity_red=values[3],
                temperature_collimator=values[4],
                humidity_collimator=values[5],
                specMech_temp=values[6],
            )

        elif this_stat == "orientation":
            write_func(accelerometer=values)

        elif this_stat == "pneumatics":
            write_func(
                shutter=values[0],
                hartmann_left=values[1],
                hartmann_right=values[2],
                air_pressure=values[3],
            )

        elif this_stat == "time":
            write_func(boot_time=values[0], clock_time=values[1], set_time=values[2])

        elif this_stat == "version":
            write_func(specMech_version=values[0])

        elif this_stat == "vacuum":
            write_func(
                vacuum_red_log10_pa=values[0],
                vacuum_blue_log10_pa=values[1],
            )

        elif this_stat == "specmech":
            write_func(
                fan=values[0],
                power_supply_volts=values[1],
            )

        elif this_stat == "nitrogen":
            write_func(
                buffer_dewar_supply_status=values[0],
                buffer_dewar_vent_status=values[1],
                red_dewar_vent_status=values[2],
                blue_dewar_vent_status=values[3],
                time_next_fill=values[4],
                max_valve_open_time=values[5],
                fill_interval=values[6],
                ln2_pressure=values[7],
                buffer_dewar_thermistor_status=values[8],
                red_dewar_thermistor_status=values[9],
                blue_dewar_thermistor_status=values[10],
            )

    return command.finish()


@mech.command()
@click.argument("POSITION", type=int, required=False)
@click.option(
    "--motor",
    type=click.Choice(["a", "b", "c"], case_sensitive=False),
    help="Move only this motor.",
)
@click.option(
    "--absolute",
    is_flag=True,
    help="Moves one motor to an absolute position.",
)
@click.option(
    "--wait/--no-wait",
    default=True,
    help="Waits until the motor has .",
)
@click.option(
    "--tolerance",
    default=2.0,
    type=float,
    help="Collimator positioning tolerance.",
)
@click.option(
    "--center",
    is_flag=True,
    help="Send the motors to their home positions.",
)
@click.option(
    "--center-position",
    default=1500.0,
    type=int,
    help="Absolute position for homing all the motors.",
)
async def move(
    command: YaoCommand,
    controllers,
    position: int | None = None,
    motor: str | None = None,
    absolute: bool = False,
    tolerance: float = 1.0,
    wait: bool = True,
    center: bool = False,
    center_position: int = 1500,
):
    """Commands the collimator motors.

    Without flags moves are relative from the current position and move all
    motors concurrently.

    """

    if not (await check_controller(command)):
        return

    if absolute is True and motor is None:
        return command.fail("--absolute requires specifying --motor.")

    if center is False and position is None:
        return command.fail("POSITION is required except with --center.")

    current = numpy.array([0, 0, 0], dtype=numpy.int16)

    minP_expected = command.actor.config["specMech"]["motors"]["minP"]
    maxP_expected = command.actor.config["specMech"]["motors"]["maxP"]

    try:
        for ii, motor_ in enumerate(["a", "b", "c"]):
            data = await command.actor.spec_mech.get_stat(f"motor-{motor_}")

            _, pos, vel, _, _, limit = data
            if vel != 0:
                raise SpecMechError(f"Motor {motor_} is moving.")
            if limit is True:
                raise SpecMechError(f"A limit switch was triggered for motor {motor_}.")
            current[ii] = pos

        for ii, motor_ in enumerate(["a", "b", "c"]):
            reply = await command.actor.spec_mech.send_data(f"r{motor_.upper()}")
            check_reply(reply)

            dmm = reply.data[3]
            if dmm[0] != "DMM":
                raise SpecMechError("Cannot retrieve DMM status.")

            minP = int(dmm[5])
            maxP = int(dmm[7])
            if minP != minP_expected or maxP != maxP_expected:
                raise SpecMechError("Min/max encoder positions out of range.")

    except SpecMechError as err:
        return command.fail(str(err))

    mech_commands: list[str]

    if center is True:
        position = center_position
        absolute = True

    if position is None:
        return command.fail("POSITION not defined.")

    if motor is None:
        if absolute is False:
            mech_commands = [f"md{position}"]
        else:
            mech_commands = [
                f"mA{position}",
                f"mB{position}",
                f"mC{position}",
            ]
    else:
        if absolute:
            mech_command = "m" + motor.upper()
        else:
            mech_command = "m" + motor.lower()
        mech_commands = [mech_command + str(position)]

    collimator_speed = 25  # mu/s

    final = current.copy()

    if motor is not None:
        motor_index = ord(motor) - ord("a")
        if absolute:
            final[motor_index] = position
        else:
            final[motor_index] += position
    else:
        if absolute:
            final[:] = position
        else:
            final[:] += position

    min_range = command.actor.config["specMech"]["motors"]["min_microns"]
    max_range = command.actor.config["specMech"]["motors"]["max_microns"]
    if numpy.any(final < min_range) or numpy.any(final > max_range):
        return command.fail("Commanded motor position is out of range.")

    move_time = numpy.max(numpy.abs(current - final)) / collimator_speed

    for mech_command in mech_commands:
        move_command = await command.actor.spec_mech.send_data(mech_command)
        if not parse_reply(command, move_command):
            return

    if not wait:
        return command.finish()

    await asyncio.sleep(move_time + 2)

    for ntry in [1, 2]:
        new = await command.actor.spec_mech.get_stat("motors")
        if numpy.allclose(new, final, atol=tolerance):
            return command.finish(motor_positions=new)

        command.warning(motor_positions=new)
        if ntry == 1:
            command.warning("Collimator not at final position. Waiting a bit longer.")
            await asyncio.sleep(3)
        else:
            return command.fail("Collimator failed to reach commanded position.")


@mech.command()
async def ack(command: YaoCommand, controllers):
    """Acknowledges the specMech has rebooted and informs the user."""

    if not (await check_controller(command)):
        return

    reply = await command.actor.spec_mech.send_data("!")

    if not parse_reply(command, reply):
        return

    if reply.code != ReplyCode.REBOOT_ACKNOWLEDGED:
        return command.fail("specMech did not acknowledge.")

    return command.finish("specMech has been acknowledged.")


@mech.command()
@click.argument("DATA", type=str)
async def talk(command: YaoCommand, controllers, data: str):
    """Send data string directly as-is to the specMech."""

    if not (await check_controller(command)):
        return

    reply = await command.actor.spec_mech.send_data(data)

    # Remove telnet negotiations from raw string.
    match = re.match(b"^(?:\xff.+\xf0)?(.*)$", reply.raw, re.DOTALL)
    if not match:
        return command.fail("Failed parsing reply from specMech.")

    return command.finish(mech_raw_reply=match.group(1).decode())


@mech.command()
@click.argument("TIME", type=str)
async def set_time(command: YaoCommand, controller, time: str):
    """Set the clock time of the specMech."""

    if not (await check_controller(command)):
        return

    dataTemp = f"st{time}"
    reply = await command.actor.spec_mech.send_data(dataTemp)

    if not parse_reply(command, reply):
        return

    return command.finish()


@mech.command(name="open")
@click.argument(
    "MECHANISMS",
    type=click.Choice(["left", "right", "shutter"], case_sensitive=False),
    nargs=-1,
    required=True,
)
async def openMech(command: YaoCommand, controller, mechanisms: tuple[str, ...]):
    """Opens left or right Hartmann doors, or the shutter."""

    if not (await check_controller(command)):
        return

    jobs = []

    for mechanism in mechanisms:
        jobs.append(
            command.actor.spec_mech.pneumatic_move(
                mechanism,
                open=True,
                command=command,
            )
        )

    results = await asyncio.gather(*jobs, return_exceptions=True)
    for result in results:
        if isinstance(result, SpecMechError):
            return command.fail(str(result))

    return command.finish()


@mech.command(name="close")
@click.argument(
    "MECHANISMS",
    type=click.Choice(["left", "right", "shutter"], case_sensitive=False),
    nargs=-1,
    required=True,
)
async def closeMech(command: YaoCommand, controller, mechanisms: tuple[str, ...]):
    """Closes left or right Hartmann doors, or the shutter."""

    if not (await check_controller(command)):
        return

    jobs = []

    for mechanism in mechanisms:
        jobs.append(
            command.actor.spec_mech.pneumatic_move(
                mechanism,
                open=False,
                command=command,
            )
        )

    results = await asyncio.gather(*jobs, return_exceptions=True)
    for result in results:
        if isinstance(result, SpecMechError):
            return command.fail(str(result))

    return command.finish()


@mech.command()
async def reboot(command: YaoCommand, controller):
    """Reboots the controller. A reconnect and acknowledge are needed afterewards."""

    if not (await check_controller(command)):
        return

    dataTemp = "R"

    reply = await command.actor.spec_mech.send_data(dataTemp)

    if not parse_reply(command, reply):
        return

    return command.finish(
        "specMech has been rebooted. You need to reconnect and acknowledge."
    )


@mech.command()
async def reconnect(command: YaoCommand, controller):
    """Recreates the connection to the specMech."""

    try:
        await command.actor.spec_mech.close()
    except OSError:
        pass

    await asyncio.sleep(2)

    try:
        await command.actor.spec_mech.start()
        if command.actor.spec_mech.writer is None:
            raise ConnectionError()
    except Exception as err:
        return command.fail(f"Failed reconnecting to the specMech: {err}")

    return command.finish("The connection to the specMech has been reestablished.")


@mech.command()
async def disconnect(command: YaoCommand, controller):
    """Closes the connection to the specMech."""

    await command.actor.spec_mech.close()

    return command.finish("The connection to the specMech has been closed.")


@mech.command()
@click.argument("MODE", type=click.Choice(["on", "off"], case_sensitive=False))
async def fan(command: YaoCommand, controller, mode: str):
    """Turns the speMech fan on/off."""

    if not (await check_controller(command)):
        return

    if mode == "on":
        mech_command = "sf+"
    else:
        mech_command = "sf-"

    reply = await command.actor.spec_mech.send_data(mech_command)

    if not parse_reply(command, reply):
        return

    return command.finish(fan="on" if mode == "on" else "off")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-11-03
# @Filename: controller.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import Callable

from archon import log
from archon.controller import ArchonController, ControllerStatus


__all__ = ["YaoController"]


class YaoController(ArchonController):
    """Extension of `.ArchonController` for Yao."""

    async def erase(self):
        """Run the LBNL erase procedure."""

        log.info(f"{self.name}: erasing.")

        await self.reset(release_timing=False, autoflush=False)

        self.update_status(ControllerStatus.FLUSHING)

        await self.set_param("DoErase", 1)
        await self.send_command("RELEASETIMING")

        await asyncio.sleep(2)  # Real time should be ~0.6 seconds.

        await self.reset()

    async def cleanup(
        self,
        erase: bool = False,
        n_cycles: int = 10,
        fast: bool = False,
        notifier: Callable[[str], None] | None = None,
    ):
        """Runs a cleanup procedure for the LBNL chip.

        Executes a number of cycles of the e-purge routine followed by a chip
        flush (complete or fast). After the e-purge cycles have been completed,
        runs three full flushing cycles.

        Parameters
        ----------
        erase
            Calls the `.erase` routine before running the e-purge cycle.
        n_cycles
            Number of e-purge/flushing cycles to execute.
        fast
            If `False`, a complete flushing is executed after each e-purge (each
            line is shifted and read). If `True`, a binning factor of 10 is used.
        notifier
            A function to call to output messages (usually a command write method).

        """

        if notifier is None:
            notifier = lambda text: None  # noqa

        if erase:
            notifier("Erasing chip.")
            await self.erase()

        mode = "fast" if fast else "normal"
        purge_msg = f"Doing {n_cycles} with DoPurge=1 (mode={mode})"
        log.info(purge_msg)
        notifier(purge_msg)

        for ii in range(n_cycles):
            notifier(f"Cycle {ii+1} of {n_cycles}.")
            await self.purge(fast=fast)

        await self.set_param("DoPurge", 0)

        flush_msg = "Flushing 3x"
        log.info(flush_msg)
        notifier(flush_msg)

        await self.flush(3)

        await self.reset()

        return True

    async def purge(self, fast: bool = True):
        """Runs a single cycle of the e-purge routine.

        A cycle consists of an execution of the e-purge routine followed by a
        chip flushing.

        Parameters
        ----------
        fast
            If `False`, a complete flushing is executed after the e-purge (each
            line is shifted and read). If `True`, a binning factor of 10 is used.

        """

        log.info("Running e-purge.")

        if fast:
            await self.set_param("FLUSHBIN", 10)
            await self.set_param("SKIPLINEBINVSHIFT", 220)
        else:
            await self.set_param("FLUSHBIN", 2200)
            await self.set_param("SKIPLINEBINVSHIFT", 1)

        await self.reset(release_timing=False)

        self.update_status(ControllerStatus.FLUSHING)

        await self.set_param("DOPURGE", 1)
        await self.send_command("RELEASETIMING")

        flush_time = self.config["timeouts"]["flushing"]
        if fast:
            flush_time = self.config["timeouts"]["fast_flushing"]

        await asyncio.sleep(self.config["timeouts"]["purge"] + flush_time)

        await self.set_param("FLUSHBIN", 2200)
        await self.set_param("SKIPLINEBINVSHIFT", 1)

        await self.reset()

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-04-01
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os
import sys

import click
from click_default_group import DefaultGroup
from clu.tools import cli_coro

from sdsstools.daemonizer import DaemonGroup

from yao import __version__
from yao.actor import YaoActor


@click.group(
    cls=DefaultGroup,
    default="actor",
    default_if_no_args=True,
    invoke_without_command=True,
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode.",
)
@click.option(
    "--version",
    is_flag=True,
    help="Print version and exit.",
)
@click.pass_context
def yao(ctx: click.Context, verbose: bool = False, version: bool = False):
    """Yao actor."""

    ctx.obj = {"verbose": verbose}

    if version is True:
        click.echo(__version__)
        sys.exit(0)


@yao.group(cls=DaemonGroup, prog="yao-actor", workdir=os.getcwd())
@click.pass_context
@cli_coro
async def actor(ctx: click.Context):
    """Runs the actor."""

    yao_actor = YaoActor.from_config()

    if ctx.obj["verbose"]:
        yao_actor.log.sh.setLevel(logging.DEBUG)

    await yao_actor.start()
    await yao_actor.run_forever()  # type: ignore


if __name__ == "__main__":
    yao()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-04-29
# @Filename: commands.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from archon.actor import parser

from .exceptions import SpecMechError


if TYPE_CHECKING:
    from .actor import YaoCommand
    from .controller import YaoController


__all__ = ["hartmann"]


async def run_subcmd(command: YaoCommand, command_string: str):
    """Runs a subcommand and checks its completion."""

    subcmd = await command.child_command(command_string)
    if subcmd.status.did_fail:
        command.fail(f"Command {command_string} failed.")
        return False

    return True


@parser.command()
@click.argument("EXPTIME", type=float)
@click.option(
    "--side",
    "-s",
    type=click.Choice(["both", "left", "right"], case_sensitive=True),
    default="both",
    help="Position of the Hartmann door. With both, takes two arcs with different "
    "doors in.",
)
@click.option(
    "--sub-frame",
    is_flag=True,
    help="Read a subregion of the frame.",
)
async def hartmann(
    command: YaoCommand,
    _,
    exptime: float,
    side: str = "both",
    sub_frame: bool = False,
):
    """Takes Hartmann image(s)."""

    spec_mech = command.actor.spec_mech

    side_list: list[str]
    if side == "both":
        side_list = ["left", "right"]
    else:
        side_list = [side]

    try:
        for ss in side_list:
            other_side = "left" if ss == "right" else "right"

            command.debug(text=f"Taking {ss} Hartmann.")

            # Check that the side door is open or open it.
            status_side = await spec_mech.pneumatic_status(ss)
            if status_side == "transitioning":
                raise SpecMechError(f"Hartmann {ss} door is transitioning.")
            elif status_side == "open":
                if not (await run_subcmd(command, f"mech close {ss}")):
                    return

            # Check that the side door is closed or close it.
            status_other_side = await spec_mech.pneumatic_status(other_side)
            if status_other_side == "transitioning":
                raise SpecMechError(f"Hartmann {other_side} door is transitioning.")
            elif status_other_side == "closed":
                if not (await run_subcmd(command, f"mech open {other_side}")):
                    return

            expose_cmd: str = "expose --arc"
            if sub_frame:
                expose_cmd += " --window-mode hartmann"
            expose_cmd += f" {exptime}"

            if not (await run_subcmd(command, expose_cmd)):
                return

        # Open closed door.
        if not (await run_subcmd(command, "mech open right")):
            return

    except SpecMechError as err:
        return command.fail(error=f"SpecMech error: {err}")

    except Exception as err:
        return command.fail(error=f"Unexpected exception: {err}")

    finally:
        # Reset default window.
        await command.send_command("yao", "set-window")

    return command.finish()


@parser.command()
async def erase(command: YaoCommand, controllers: dict[str, YaoController]):
    """Runs the r2 erase routine."""

    for controller in controllers.values():
        command.info(f"Running erase routine on spectrograph {controller.name}.")
        await controller.erase()

    return command.finish("All done.")


@parser.command()
@click.option("--erase", is_flag=True, help="Run the erase procedure too.")
@click.option("--cycles", type=int, default=5, help="Number of purge cycles.")
@click.option("--slow", is_flag=True, help="Does full flushing for each cycle.")
async def cleanup(
    command: YaoCommand,
    controllers: dict[str, YaoController],
    erase: bool = False,
    cycles: int = 5,
    slow: bool = False,
):
    """Runs the r2 cleanup routine."""

    for controller in controllers.values():
        command.info(f"Running cleanup routine on spectrograph {controller.name}.")
        await controller.cleanup(
            erase=erase,
            notifier=command.info,
            n_cycles=cycles,
            fast=not slow,
        )

    return command.finish("All done.")


@parser.group()
def purge(*args):
    """Sets the purge routine on/off."""

    pass


@purge.command()
async def on(command: YaoCommand, controllers: dict[str, YaoController]):
    """Sets DoPurge=1."""

    for controller in controllers.values():
        await controller.set_param("DoPurge", 1)

    return command.finish("All done.")


@purge.command()
async def off(command: YaoCommand, controllers: dict[str, YaoController]):
    """Sets DoPurge=0."""

    for controller in controllers.values():
        await controller.set_param("DoPurge", 0)

    return command.finish("All done.")



# encoding: utf-8

from sdsstools import get_config, get_package_version


# pip package name
NAME = "sdss-yao"

# Loads config. config name is the package name.
config = get_config("yao")


# package name should be pip package name
__version__ = get_package_version(path=__file__, package_name=NAME)


from .actor import YaoActor, YaoCommand
from .delegate import YaoDelegate
from .mech_commands import *
from .mech_controller import MechController
